'''
portScanner.py

Copyright 2006 Andres Riancho

This file is part of w3af, w3af.sourceforge.net .

w3af is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation version 2 of the License.

w3af is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with w3af; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

'''

from core.controllers.basePlugin.baseDiscoveryPlugin import baseDiscoveryPlugin
import core.controllers.outputManager as om
from core.data.parsers.urlParser import getDomain
from core.controllers.w3afException import w3afException
import core.data.kb.knowledgeBase as kb
import core.data.kb.info as info

class portScanner(baseDiscoveryPlugin):
	'''
	This plugin does a port scan against the server that hosts the site in order to find other HTTP servers.
	
	@author: Andres Riancho ( andres.riancho@gmail.com )
	'''

	def __init__(self):
		baseDiscoveryPlugin.__init__(self)
		
		self._portRange = '1:1024'
		self._portRangeTuple = self._parseRange( self._portRange )

	def discover(self, fuzzableRequest ):
		'''
		Portscans the remote server.
		
		@parameter fuzzableRequest: A fuzzableRequest instance that contains (among other things) the URL to test.
		'''
		om.out.debug( 'portScanner plugin is testing: ' + fuzzableRequest.getURL() )
		
		# Init some internal variables
		self._results = []
		
		domain = getDomain( fuzzableRequest.getURL() )
		
		for port in range( self._portRangeTuple[0], self._portRangeTuple[1] ) :
			targs = ('http://' + domain + ':' + str(port) + '/', )
			self._tm.startFunction( target=self._doGet, args=targs, ownerObj=self )

			targs = ('https://' + domain + ':' + str(port) + '/', )
			self._tm.startFunction( target=self._doGet, args=targs, ownerObj=self )
			
		self._tm.join( self )
		
		if len( self._results ):
			om.out.information('The portscanner plugin found the following HTTP(s) servers:')
			for s in self._results:
				om.out.information('- ' + s)
				i = info.info()
				i.setURL( s )
				i.setDesc('The remote server has a webserver at ' + s)
				kb.kb.append( self, 'ports', i )
		else:
			om.out.information('The portscanner plugin failed to found HTTP(s) servers.')
		
		return []
	
	def _doGet( self, url ):
		'''
		Does a GET request and stores successfull results to self._results
		'''
		try:
			res = self._urlOpener.GET( url, useCache=True )
		except w3afException, w3:
			#om.out.debug( str(w3) )
			return
		else:
			self._results.append( url )
		
	
	def getOptionsXML(self):
		'''
		This method returns a XML containing the Options that the plugin has.
		Using this XML the framework will build a window, a menu, or some other input method to retrieve
		the info from the user. The XML has to validate against the xml schema file located at :
		w3af/core/ui/userInterface.dtd
		
		@return: XML with the plugin options.
		'''	
		return	'<?xml version="1.0" encoding="ISO-8859-1"?>\
		<OptionList>\
			<Option name="portRange">\
				<default>'+self._portRange+'</default>\
				<desc>Port range to scan</desc>\
				<type>string</type>\
			</Option>\
		</OptionList>\
		'

	def setOptions( self, optionsMap ):
		'''
		This method sets all the options that are configured using the user interface 
		generated by the framework using the result of getOptionsXML().
		
		@parameter OptionList: A dictionary with the options for the plugin.
		@return: No value is returned.
		'''	
		self._portRangeTuple = self._parseRange( optionsMap['portRange'] )
		self._portRange = optionsMap['portRange']
		
	def _parseRange( self, rangeStr ):
		try:
			a , b = rangeStr.split(':')
		except:
			raise w3afException('Invalid range, sintax should be: "a:b"')
		else:
			try:
				a = int(a)
				b = int(b)
			except:
				raise w3afException('Invalid range, the range must be specified by integers.')
			else:
				if a > 0 and b <= 65535:
					return a, b
				else:
					raise w3afException('Invalid range, the range should be "a:b" with a>0 and b<=65535')
			
		
	def getPluginDeps( self ):
		'''
		@return: A list with the names of the plugins that should be runned before the
		current one.
		'''
		return [ ]
			
	def getLongDesc( self ):
		'''
		@return: A DETAILED description of the plugin functions and features.
		'''
		return '''
		This plugin does a port scan against the server that hosts the site in order to find other HTTP servers.
		Because w3af uses xUrllib to do the scan connections, this port scanner has no service fingerprint, the gain 
		is that the plugin uses the configured proxy settings.
		
		There is one configurable parameter:
			- portRange
			
		The syntaxis for port ranges is : 'a:b' where a > 0 and b <= 65535.
		'''
