'''
userDir.py

Copyright 2006 Andres Riancho

This file is part of w3af, w3af.sourceforge.net .

w3af is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation version 2 of the License.

w3af is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with w3af; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

'''

import core.controllers.outputManager as om
import core.data.parsers.urlParser as urlParser
from core.controllers.basePlugin.baseDiscoveryPlugin import baseDiscoveryPlugin
import core.data.kb.knowledgeBase as kb
from core.controllers.w3afException import w3afException
from core.controllers.w3afException import w3afRunOnce

class userDir(baseDiscoveryPlugin):
	'''
	This plugin tries to find user directories ( http://test/~user/ ) .
	@author: Andres Riancho ( andres.riancho@gmail.com )  
	'''

	def __init__(self):
		baseDiscoveryPlugin.__init__(self)
		self._run = True
		
	def discover(self, fuzzableRequest ):
		'''
		Searches for user directories.
		
		@parameter fuzzableRequest: A fuzzableRequest instance that contains (among other things) the URL to test.
		'''
		if not self._run:
			raise w3afRunOnce()
			
		else:
			self._run = False
			self._fuzzableRequests = []
				
			url = urlParser.baseUrl( fuzzableRequest.getURL() )
			self._headers = {'Referer':url }
			self.is404 = kb.kb.getData( 'error404page', '404' )
			
			# Create a response body to compare with the others
			nonExistantUser = '~_w_3_a_f_/'
			testURL = urlParser.urlJoin( urlParser.baseUrl( fuzzableRequest.getURL() ),  nonExistantUser )
			try:
				responseBody = self._urlOpener.GET( testURL, useCache=True, headers=self._headers ).getBody()
			except:
				raise w3afException('userDir failed to create a non existant signature.')
			self._nonExistant = responseBody.replace( nonExistantUser, '')
			
			# Check the users to see if they exist
			mutantURLList = self._createDirs( url )
			for mutant in mutantURLList :
				om.out.debug('userDir is testing ' + url )
				targs = ( mutant, )
				self._tm.startFunction( target=self._doRequest, args=targs, ownerObj=self )
			self._tm.join( self )
			
			return self._fuzzableRequests

	def _doRequest( self, mutant ):
		try:
			response = self._urlOpener.GET( mutant, useCache=True, headers=self._headers )
		except KeyboardInterrupt,e:
			raise e
		else:
			path = mutant.replace( urlParser.baseUrl( mutant ) , '' )
			responseBody = response.getBody().replace( path, '')
			if responseBody != self._nonExistant:
				om.out.information('A user directory was found at: ' + mutant )
				frList = self._createFuzzableRequests( response )
				self._fuzzableRequests.extend( frList )
	
	def _createDirs(self, url):
		'''
		Append the users to the URL.
		
		@param url: The original url
		@return: A list of URL's with the username appended.
		'''
		res = []
		for user in self._getUsers():
			res.append( urlParser.urlJoin( url , '/'+user+'/' ) )
			res.append( urlParser.urlJoin( url , '/~'+user+'/' ) )
		return res
	
	def _getUsers( self ):
		'''
		@return: All usernames collected by other plugins.
		'''
		res = []
		
		infoList = kb.kb.getData( 'mails', 'mails' )
		
		for i in infoList:
			res.append( i['user'] )
		
		# Add some common users:
		res.extend( ['www-data', 'www', 'nobody', 'root' , 'admin' , 'test'] )
			
		return res
		
	
	def getOptionsXML(self):
		'''
		This method returns a XML containing the Options that the plugin has.
		Using this XML the framework will build a window, a menu, or some other input method to retrieve
		the info from the user. The XML has to validate against the xml schema file located at :
		w3af/core/ui/userInterface.dtd
		
		@return: XML with the plugin options.
		'''	
		return	'<?xml version="1.0" encoding="ISO-8859-1"?>\
		<OptionList>\
		</OptionList>\
		'

	def setOptions( self, optionList ):
		'''
		This method sets all the options that are configured using the user interface 
		generated by the framework using the result of getOptionsXML().
		
		@parameter optionList: A dictionary with the options for the plugin.
		@return: No value is returned.
		'''	
		pass
	
	def getPluginDeps( self ):
		'''
		@return: A list with the names of the plugins that should be runned before the
		current one.
		'''
		# This was left here for fast testing of the plugin.
		#return ['discovery.error404page','discovery.fingerPKS' ]
		
		# This is the correct return value for this method.
		return ['discovery.error404page','discovery.fingerMSN', 'discovery.fingerGoogle', 'discovery.fingerPKS' ]
	
	def getLongDesc( self ):
		'''
		@return: A DETAILED description of the plugin functions and features.
		'''
		return '''
		This plugin will try to find user home directories based on the knowledge gained by other plugins.
		For example, if the target URL is:
			- http://test/
			
		And other plugins found this valid email accounts:
			- test@test.com
			- f00b4r@test.com
			
		This plugin will request:
			- http://test/~test/
			- http://test/test/
			- http://test/~f00b4r/
			- http://test/f00b4r/
		
		If the response is not a 404 error, then we have found a new URL. And confirmed the existance of a user
		in the remote system.
		'''
